# pylint: disable=superfluous-parens
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# pylint: disable=attribute-defined-outside-init

"""
A library which provides helper functions and classes.
"""

import logging
import logging.handlers
import os
import sys
import socket
import json
import daemon
import requests

from anycast_healthchecker import __version__ as version


class LoggingDaemonContext(daemon.DaemonContext):
    """A subclass of daemom.DaemonContext to add support for loggers.

    Arguments:
        loggers_preserve (list): A list of loggers.
        stdout_logger (logger obj): A logger for stdout.
        stderr_logger (logger obj): A logger for stderr.

    Methods:
        open(): Overwrites open method of daemon.DaemonContext.
    """
    def _add_logger_files(self):
        """Adds all files related to loggers into files_preserve."""
        for logger in [self.stdout_logger, self.stderr_logger]:
            if logger:
                self.loggers_preserve.append(logger)
        logger_files = open_files_from_loggers(self.loggers_preserve)
        self.files_preserve.extend(logger_files)

    def __init__(self,
                 chroot_directory=None,
                 working_directory='/',
                 umask=0o022,
                 uid=None,
                 gid=None,
                 prevent_core=True,
                 detach_process=None,
                 files_preserve=None,   # changed default
                 loggers_preserve=None,  # new
                 pidfile=None,
                 stdout_logger=None,  # new
                 stderr_logger=None,  # new
                 # stdin,   omitted!
                 # stdout,  omitted!
                 # sterr,   omitted!
                 signal_map=None):

        self.stdout_logger = stdout_logger
        self.stderr_logger = stderr_logger
        self.loggers_preserve = loggers_preserve

        devnull_in = open(os.devnull, 'r+')
        devnull_out = open(os.devnull, 'w+')
        if files_preserve is not None:
            files_preserve.extend([devnull_in, devnull_out])
        else:
            files_preserve = [devnull_in, devnull_out]

        super(LoggingDaemonContext, self).__init__(
            chroot_directory=chroot_directory,
            working_directory=working_directory,
            umask=umask,
            uid=uid,
            gid=gid,
            prevent_core=prevent_core,
            detach_process=detach_process,
            files_preserve=files_preserve,
            pidfile=pidfile,
            stdin=devnull_in,
            stdout=devnull_out,
            stderr=devnull_out,
            signal_map=signal_map)
        # python-daemon>=2.1 has initgroups=True by default which requires
        # root privileges(CAP_SETGID capability).
        # Older versions don't have it, so we set it manually instead of
        # passing it to the supper()
        self.initgroups = False

    def open(self):
        """Redirects stdout/stderr to loggers and calls DaemonContext.open."""
        self._add_logger_files()
        daemon.DaemonContext.open(self)
        if self.stdout_logger:
            sys.stdout = self.stdout_logger
        if self.stderr_logger:
            sys.stderr = self.stderr_logger


def open_files_from_loggers(loggers):
    """Returns open files used by file-based logger handlers.

    Arguments:
        loggers (list): A list of logger objects.

    Returns:
        A list of open files used by file-based logger handlers.
    """
    open_files = []
    for logger in loggers:
        for handler in logger.logger.handlers:
            if hasattr(handler, 'stream') and \
               hasattr(handler.stream, 'fileno'):
                open_files.append(handler.stream)
    return open_files


class LoggerExt(object):
    """Create a logging.Logger class with extended functionality

    It wraps a Logger class into a file like object, which provides a handy
    way to redirect stdout/stdin to a logger. It also accepts a JSON blob as
    input and forwards it over HTTP to a central location. The JSON blob is
    built out of a certain data structure which isn't configurable. This
    provides a easy way to pass messages to an ElasticSearch infrastructure.

    Arguments
        name (str): The name for the logger.
        file_path (str): The absolute path of the log file.
        log_level (logging.level obj): The threshold for this logger.
        log_format (logging.Formatter): The format of this logger.
        maxbytes (int): Max size of the log before it is rotated.
        backupcount (int): Number of backup file to keep.


    Methods:
        write(string): Writes string to logger with newlines removed.
        flush(): Flushes logger messages.
        close(): Closes logger.

    Returns:
        A logger object.

    Note:
        See logging module for acceptable values for log_level
        and log_format.
    """
    def __init__(self,
                 name,
                 file_path,
                 log_level=logging.DEBUG,
                 log_format=None,
                 maxbytes=104857600,
                 backupcount=8,):

        if log_format is None:
            log_format = ('%(asctime)s [%(process)d] %(levelname)-8s '
                          '%(threadName)-32s %(message)s')
        self.logger = logging.getLogger(name)
        self.logger.setLevel(log_level)
        handler = logging.handlers.RotatingFileHandler(file_path,
                                                       maxBytes=maxbytes,
                                                       backupCount=backupcount)
        formatter = logging.Formatter(log_format)
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

        self.jid = "anycast-healthchecker-{h}".format(
            h=socket.gethostname().split('.')[0])

    def __getattr__(self, name):
        """Return a logger function for emitting messages

        Because it acts as a proxy for all undefined attributes, we only
        allow the ones that we know Logger will accept.
        """
        _valid_methods = [
            'critical',
            'warning',
            'warn',
            'info',
            'notice',
            'debug',
            'error',
        ]
        if name in _valid_methods:
            def log(msg, priority=10, json_blob=True, **kwargs):
                """A wrapper around logger method

                It extends the capabilities by sending also the messages to
                a HTTP server.

                Arguments:
                    msg(string): A message to emit
                    priority(integer): The priority associated with the
                    messsage
                    json_blob(boolean): ``True`` to send a JSON blob
                    kwargs(dictinary): A dictionary with extra information to
                    add to the JSON blob

                Returns:
                    A logger function
                """
                _logger = getattr(self.logger, name)
                _logger(msg)

                # send msg over http only if we are configured
                if hasattr(self, 'server') and json_blob:
                    self._send_http(msg, priority, **kwargs)

            return log
        else:
            raise AttributeError

    def write(self, string):
        """Erases newline from a string and writes to the logger."""
        string = string.rstrip()
        if string:  # Don't log empty lines
            if hasattr(self, 'server'):
                self._send_http(string, priority=80)
            for line in string.split('\n'):
                # Critical to log at any log level.
                self.logger.critical(line)

    def flush(self):
        """Flushes logger's data."""
        # In case multiple handlers are attached to the logger make sure
        # they are flushed.
        for handler in self.logger.handlers:
            handler.flush()

    def close(self):
        """Calls the closer method of the logger."""
        for handler in self.logger.handlers:
            handler.close()

    def add_central_logging(self,
                            server='127.0.0.1',
                            timeout=1,
                            protocol='http',
                            port=2813,
                            path='/'):
        """Extends logger to a HTTP client"""
        self.server = server
        self.timeout = timeout
        self.protocol = protocol
        self.port = port
        self.path = path
        # Use Session object from requests for enforcing HTTP persistence
        # connection (HTTP keep-alive) so the underlying TCP connection will be
        # reused and recycled. This reduces the stress on the HTTP entry point.
        self._http_sess = requests.Session()

    def _send_http(self, msg, priority=10, **kwargs):
        """Send msg as a JSON blob"""
        # These are the mandatory elements of the data structure which we send
        kwargs['softwareversion'] = version
        data = {
            'id': self.jid,
            'msg_type': 'anycast-healthchecker',
            'priority': priority,
            'msg_text': msg,
            'extra': kwargs,
        }
        # since we send JSON make sure we set Content-Type accordigly
        headers = {
            'Content-type': 'application/json',
            'Accept': 'text/plain',
        }
        url = "{proto}://{host}:{port}{path}".format(proto=self.protocol,
                                                     host=self.server,
                                                     port=self.port,
                                                     path=self.path)
        try:
            req = self._http_sess.post(url, timeout=self.timeout,
                                       data=json.dumps(data), headers=headers)
        except (requests.exceptions.Timeout,
                requests.exceptions.ConnectionError,
                requests.exceptions.RequestException) as error:
            self.logger.warning("failed to send data to %s: %s",
                                self.server, error)
        else:
            if req.status_code == 200:
                try:
                    response = req.json()
                except ValueError as error:
                    self.logger.warning("failed to decode JSON response (%s)"
                                        ": %s", req.text, error)
                else:
                    # a valid response(JSON) looks like
                    # {"Status":"OK","Responses":["OK"]}
                    if (response['Status'] != 'OK' and
                            'OK' not in response['Responses']):
                        self.logger.warning("something went wrong when we sent"
                                            " (%s) as response from server "
                                            "was (%s)", data, req.text)
            else:
                self.logger.warning("failed to send JSON data, received HTTP "
                                    "status code %s with response content "
                                    "(%s)", req.status_code, req.text)



# pylint: disable=superfluous-parens
# pylint: disable=too-many-arguments

"""
A library which provides the HealthChecker class.
"""
import os
import subprocess
import sys
import time
import datetime
from queue import Queue
import shlex

from anycast_healthchecker.servicecheck import ServiceCheck
from anycast_healthchecker.utils import (OPTIONS_TYPE,
                                         get_ip_prefixes_from_bird,
                                         ip_prefixes_without_config)


class HealthChecker(object):
    """Lunches service checks and triggers a reconfiguration on BIRD daemon.

    This class should be instantiated once and daemonized.

    It uses a queue as a store for IP prefixes to be removed from and added to
    BIRD configuration file.

    Arguments:
        log(logger obj): A logger to log messages.

        config(configparger obj): A configparser object with the configuration

        bird_conf_file('str'): The absolute path of file which contains the
        definition of constant used by BIRD daemon to store IP prefixes for
        which routes are allowed to be advertised

        bird_constant_name('str'): The constant name used in the BIRD daemon
        configuration.

        dummy_prefix('str'): A dummy IP prefix which must be always present
        in bird_constant_name and it should never be removed from it.

        action(Queue obj): A queue of IP prefixes and their action to be taken
        based on the state of health check. An item is a tuple of 3 elements:
            1st: name of the thread.
            2nd: IP prefix.
            3nd: Action to take, either 'add' or 'del'.

    Methods:
        run(): Lunches checks and updates BIRD configuration based on
        the result of the check.
        catch_signal(signum, frame): Catches signals
    """
    def __init__(self,
                 log,
                 config,
                 bird_conf_file,
                 bird_constant_name,
                 dummy_ip_prefix):
        self.log = log
        self.config = config
        self.bird_conf_file = bird_conf_file
        self.dummy_ip_prefix = dummy_ip_prefix
        self.bird_constant_name = bird_constant_name
        self.action = Queue()

        # A list of service of checks
        self.services = config.sections()
        self.services.remove('daemon')

        self.log.info('initialize healthchecker')

    def _write_temp_bird_conf(self, prefixes):
        """Writes in a temporary file the list of IP-Prefixes

        A failure to create and write the temporary file will exit main
        program.

        Arguments:
            prefixes (list): The list of IP-Prefixes to write
        Returns:
            The filename of the temporary file
        """
        comment = ("# {i} is a dummy IP Prefix. It should NOT be used and "
                   "REMOVED from the constant.".format(i=self.dummy_ip_prefix))

        # the temporary file must be on the same filesystem as the bird config
        # as we use os.rename to perform an atomic update on the bird config.
        # Thus, we create it in the same directory that bird config is stored.
        tm_file = os.path.dirname(self.bird_conf_file) + '/' + str(time.time())
        self.log.debug("going to write to {f}".format(f=tm_file),
                       json_blob=False)
        try:
            with open(tm_file, 'w') as tmpf:
                tmpf.write("# Generated {t} by anycast-healthchecker (pid={p})"
                           "\n".format(t=datetime.datetime.now(),
                                       p=os.getpid()))
                tmpf.write("{c}\n".format(c=comment))
                tmpf.write("define {n} =\n".format(n=self.bird_constant_name))
                tmpf.write("{s}[\n".format(s=4 * ' '))
                # all entries of the array constant need a trailing comma
                # except the last one. A single element array doesn't need
                # the trailing comma.
                tmpf.write(',\n'.join([' '*8 + n for n in prefixes]))
                tmpf.write("\n{s}];\n".format(s=4 * ' '))
        except OSError as error:
            msg = ("failed to write temporary file {f}: {e}. This is a FATAL "
                   "error, this exiting main program".format(f=tm_file,
                                                             e=error))
            self.log.critical(msg, priority=80)
            sys.exit(1)
        else:
            return tm_file

    def _update_bird_conf_file(self, operation):
        """Updates BIRD configuration.

        Adds/removes entries from a list and updates generation time stamp.
        Main program will exit if configuration file cant be read/written.

        Arguments:
            operation (obj): Either an AddOperation or DeleteOperation object

        Returns:
            True if BIRD configuration was updated otherwise False.

        """
        conf_updated = False
        prefixes = []

        try:
            prefixes = get_ip_prefixes_from_bird(self.bird_conf_file,
                                                 die=False)
        except OSError as error:
            msg = ('failed to open Bird configuration {e}, this is a FATAL'
                   'error, thus exiting main program').format(e=error)
            self.log.error(msg, priority=80)
            sys.exit(1)

        if not prefixes:
            msg = ("found empty bird configuration:{f}, this is a FATAL"
                   "error, thus exiting main program").format(
                       f=self.bird_conf_file)
            self.log.error(msg, priority=80)
            sys.exit(1)

        if self.dummy_ip_prefix not in prefixes:
            msg = ("dummy IP prefix {i} wasn't found in bird configuration, "
                   "adding it. This shouldn't have happened!").format(
                       i=self.dummy_ip_prefix)
            self.log.warning(msg, priority=20)
            prefixes.insert(0, self.dummy_ip_prefix)
            conf_updated = True

        # Remove IP prefixes for which we don't have a configuration for them.
        notconfigured_ip_prefixes = ip_prefixes_without_config(prefixes,
                                                               self.config,
                                                               self.services)
        if notconfigured_ip_prefixes:
            msg = ("found {i} IP prefixes in Bird configuration but we aren't "
                   "configured to run health checks on them. Either someone "
                   "modified the configuration manually or something went "
                   "horrible wrong. Thus, we remove them from Bird "
                   "configuration").format(
                       i=','.join(notconfigured_ip_prefixes))
            self.log.warning(msg, priority=20)
            # This is faster than using lambda and filter.
            # NOTE: We don't use remove method as we want to remove more than
            # occurrences of the IP prefixes without check.
            prefixes[:] = (ip for ip in prefixes
                           if ip not in notconfigured_ip_prefixes)
            conf_updated = True

        # Update the list of IP prefixes based on the status of health check.
        if operation.update(prefixes):
            conf_updated = True

        if not conf_updated:
            self.log.info('no updates for bird configuration')
            return conf_updated

        # some IP prefixes are either removed or added, create
        # configuration with new data.
        tempname = self._write_temp_bird_conf(prefixes)
        try:
            os.rename(tempname, self.bird_conf_file)
        except OSError as error:
            msg = ('failed to create Bird configuration {e}, this is a FATAL '
                   'error, thus exiting main program').format(e=error)
            self.log.critical(msg, priority=80)
            sys.exit(1)
        else:
            self.log.info('Bird configuration is updated')

        # dummy_ip_prefix is always there
        if len(prefixes) == 1:
            self.log.warning("Bird configuration doesn't have IP prefixes for "
                             "any of the services we monitor! It means local "
                             "node doesn't receive any traffic", priority=80)

        return conf_updated

    def _reload_bird(self):
        """Reloads BIRD daemon.

        Runs 'birdc configure' to reload BIRD. Some useful information on how
        birdc tool works:
            -- Returns a non-zero exit code only when it can't access BIRD
            daemon via the control socket (/var/run/bird.ctl).
            This happens when BIRD daemon is either down or when the caller of
            birdc doesn't have access to the control socket.
            -- Returns zero exit code when reload fails due to invalid
            configuration. Thus, we catch this case by looking at the output
            and not at the exit code.
            -- Returns zero exit code when reload was successful.
            -- Should never timeout, if it does then it is a bug.
        """
        _cmd = shlex.split(self.config['daemon']['bird_reconfigure_cmd'])
        try:
            _output = subprocess.check_output(
                _cmd,
                timeout=2,
                stderr=subprocess.STDOUT,
                universal_newlines=True)
        except subprocess.TimeoutExpired:
            self.log.error("reloading bird timed out", priority=80)
            return
        except subprocess.CalledProcessError as error:
            # birdc returns 0 even when it fails due to invalid config,
            # but it returns 1 when BIRD is down.
            msg = ("reloading BIRD failed, most likely BIRD daemon is down"
                   ":{e}").format(e=error.output.strip())
            self.log.error(msg, priority=80)
            return
        except FileNotFoundError as error:
            msg = "Reloading BIRD failed with: {e}".format(e=error)
            self.log.error(msg, priority=80)
            return

        # 'Reconfigured' string will be in the output if and only if conf is
        # valid.
        if 'Reconfigured' in _output:
            self.log.info('Reloaded BIRD daemon')
        else:
            # We will end up here only if we generated an invalid conf
            # or someone broke bird.conf.
            msg = ("reloading BIRD returned error, most likely we generated "
                   "an invalid configuration file or Bird configuration in "
                   "general is broken:{e}").format(e=_output)
            self.log.error(msg, priority=80)

    def run(self):
        """Lunches checks and triggers updates on BIRD configuration."""

        # Lunch a thread for each configuration
        msg = "going to lunch {n} threads".format(n=len(self.services))
        self.log.info(msg)
        for service in self.services:
            msg = "lunching thread for {n}".format(n=service)
            self.log.debug(msg, json_blob=False)
            _config = {}
            for option, getter in OPTIONS_TYPE.items():
                _config[option] = getattr(self.config, getter)(service, option)
            _thread = ServiceCheck(
                service,
                _config,
                self.action,
                self.log)
            _thread.start()

        # Stay running until we are stopped
        while True:
            # Fetch items from action queue
            operation = self.action.get(block=True)
            msg = ("returned an item from the queue for {n} with IP prefix {i}"
                   " and action to {o} Bird configuration").format(
                       n=operation.name,
                       i=operation.ip_prefix,
                       o=operation)
            self.log.info(msg)
            bird_updated = self._update_bird_conf_file(operation)
            self.action.task_done()
            if bird_updated:
                self._reload_bird()

    def catch_signal(self, signum, frame):
        """A signal catcher.

        Arguments:
            signum (int): The signal number.
            frame (str): The stack frame at the time the signal was received.
        """
        self.log.info("received {n} signal {f}".format(n=signum,
                                                       f=frame), priority=80)
        self.log.info('going down', priority=50)
        sys.exit(0)



# pylint: disable=superfluous-parens
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
#

"""
A library which provides the ServiceCheck class.
"""

import subprocess
import time
from threading import Thread
import shlex

from anycast_healthchecker.utils import AddOperation, DeleteOperation


class ServiceCheck(Thread):
    """Handles the health checking for a service.

    Arguments:
        service (str): The name of the service to monitor.
        config (dict): A dictionary with the configuration of the service.
        action (Queue obj): A queue object to place actions based on the result
        of the health check.
        log (logger obj): A logger object to use.

    Methods:
        run(): Run method of the thread.
    """
    def __init__(self, service, config, action, log):
        """Set the name of thread to be the name of the service."""
        super(ServiceCheck, self).__init__()
        self.name = service
        self.daemon = True
        self.config = config
        self.action = action
        self.log = log
        self.log.info("loading check for {n}".format(n=self.name))
        self.extra = {
            'servicename': self.name,
        }

    def _run_check(self):
        """Executes a check command.

        Returns:
            True if the exit code of the command was 0 otherwise False.
        """
        cmd = shlex.split(self.config['check_cmd'])
        self.log.info("running {}".format(' '.join(cmd)), **self.extra)
        proc = subprocess.Popen(cmd,
                                stdin=None,
                                stdout=None,
                                stderr=None,
                                close_fds=True)

        start_time = time.time()
        try:
            proc.wait(self.config['check_timeout'])
        except subprocess.TimeoutExpired:
            self.log.error("check timed out", priority=80, **self.extra)
            if proc.poll() is None:
                try:
                    proc.kill()
                except PermissionError:
                    self.log.warning("failed to kill check due to adequate "
                                     "access rights, check could be running "
                                     "under another user(root) via sudo",
                                     priority=80, **self.extra)

            return False
        else:
            msg = "check duration {t:.3f}ms".format(
                t=(time.time() - start_time) * 1000)
            self.log.info(msg, **self.extra)
            return proc.returncode == 0

    def _ip_assigned(self):
        """Checks if IP prefix is assigned to loopback interface.

        Returns:
            True if IP prefix found assigned otherwise False.
        """
        cmd = [
            '/sbin/ip',
            'address',
            'show',
            'dev',
            self.config['interface'],
            'to',
            self.config['ip_prefix'],
        ]

        self.log.debug("running {}".format(' '.join(cmd)), json_blob=False)
        try:
            out = subprocess.check_output(
                cmd,
                universal_newlines=True,
                timeout=1)
        except subprocess.CalledProcessError as error:
            msg = "error checking IP-PREFIX {c} {e}".format(c=cmd,
                                                            e=error.output)
            self.log.error(msg, priority=60, **self.extra)
            # Because it is unlikely to ever get an error I return True
            return True
        except subprocess.TimeoutExpired:
            msg = "timeout running {c}".format(c=' '.join(cmd))
            self.log.error(msg, priority=50, **self.extra)
            # Because it is unlikely to ever get a timeout I return True
            return True
        except ValueError as error:
            # We have been getting intermittent ValueErrors, see here
            # gist.github.com/unixsurfer/67db620d87f667423f6f6e3a04e0bff5
            # It has happened ~5 times and this code is executed from multiple
            # threads and every ~10secs on several(~40) production servers for
            # more than 18months.
            # It could be a bug in Python or system returns corrupted data.
            # As a consequence of the raised exception thread dies and the
            # service isn't monitored anymore!.
            # In order to avoid this we catch the exception.
            # While checking if IP is assigned, we get an error unrelated to
            # the IP which prevents  us from knowing if it's assigned.
            # We simply don't know.
            # A retry logic icould be a more proper solution.
            msg = ("running {c} raised ValueError exception:{e}"
                   .format(c=' '.join(cmd), e=error))
            self.log.error(msg, priority=50, **self.extra)
            return True
        else:
            if self.config['ip_prefix'] in out:  # pylint: disable=E1135
                msg = "{i} assigned to loopback interface".format(
                    i=self.config['ip_prefix'])
                self.log.debug(msg, json_blob=False)
                return True
            else:
                msg = "{i} NOT assigned to loopback interface".format(
                    i=self.config['ip_prefix'])
                self.log.warning(msg, priority=20, **self.extra)
                return False

        self.log.debug("I shouldn't land here!, it is a BUG",
                       priority=50, **self.extra)

        return False

    def _check_disabled(self):
        """Checks if health check is disabled.

        It logs a message if health check is disabled and it also adds an item
        to the action queue based on 'on_disabled' setting.

        Returns:
            True if check is disabled otherwise False.
        """
        if (self.config['check_disabled'] and
                self.config['on_disabled'] == 'withdraw'):
            self.log.info("Check is disabled and ip_prefix will be withdrawn",
                          priority=20, **self.extra)
            del_operation = DeleteOperation(name=self.name,
                                            ip_prefix=self.config['ip_prefix'],
                                            log=self.log, **self.extra)
            msg = "{i} added in queue".format(i=self.config['ip_prefix'])
            self.log.info(msg, **self.extra)
            self.action.put(del_operation)
            self.log.info("Check is now permanently disabled",
                          priority=20, status='disabled', **self.extra)
            return True
        elif (self.config['check_disabled'] and
              self.config['on_disabled'] == 'advertise'):
            self.log.info("check is disabled, ip_prefix wont be withdrawn",
                          priority=80, **self.extra)
            add_operation = AddOperation(name=self.name,
                                         ip_prefix=self.config['ip_prefix'],
                                         log=self.log, **self.extra)
            msg = "{i} add in queue".format(i=self.config['ip_prefix'])
            self.log.info(msg, **self.extra)
            self.action.put(add_operation)
            self.log.info('check is now permanently disabled',
                          priority=20, status='disabled', **self.extra)
            return True

        return False

    def run(self):
        """Discovers the health of a service.

        Runs until it is being killed from main program and is responsible to
        put an item into the queue based on the status of the health check.
        The status of service is consider UP after a number of consecutive
        successful health checks, in that case it asks main program to add the
        IP prefix associated with service to BIRD configuration, otherwise ask
        for a removal.
        Rise and fail options prevent unnecessary configuration changes when
        check is flapping.
        """
        up_cnt = 0
        down_cnt = 0
        # The current established state of the service check, it can be
        # either UP or DOWN but only after a number of consecutive successful
        # or unsuccessful health checks.
        check_state = 'Unknown'

        for key, value in self.config.items():
            self.log.debug("{}={}:{}".format(key, value, type(value)),
                           json_blob=False)

        # Service check will abort if it is disabled.
        if self._check_disabled():
            return

        interval = self.config['check_interval']
        start_offset = time.time() % interval
        # Go in a loop until we are told to stop
        while True:
            timestamp = time.time()
            if not self._ip_assigned():
                up_cnt = 0
                msg = ("status DOWN because {i} isn't assigned to loopback "
                       "interface."
                       .format(i=self.config['ip_prefix']))
                self.log.warning(msg, priority=80, status='down', **self.extra)
                if check_state != 'DOWN':
                    check_state = 'DOWN'
                    del_operation = DeleteOperation(
                        name=self.name,
                        ip_prefix=self.config['ip_prefix'],
                        log=self.log, **self.extra)
                    msg = "{i} in queue".format(i=self.config['ip_prefix'])
                    self.log.info(msg, **self.extra)
                    self.action.put(del_operation)
            elif self._run_check():
                if up_cnt == (self.config['check_rise'] - 1):
                    self.log.info("status UP", status='up', **self.extra)
                    # Service exceeded all consecutive checks. Set its state
                    # accordingly and put an item in queue. But do it only if
                    # previous state was different, to prevent unnecessary bird
                    # reloads when a service flaps between states.
                    if check_state != 'UP':
                        check_state = 'UP'
                        operation = AddOperation(
                            name=self.name,
                            ip_prefix=self.config['ip_prefix'],
                            log=self.log, **self.extra)
                        msg = "{i} in queue".format(i=self.config['ip_prefix'])
                        self.log.info(msg, **self.extra)
                        self.action.put(operation)
                elif up_cnt < self.config['check_rise']:
                    up_cnt += 1
                    msg = "going up {n}".format(n=up_cnt)
                    self.log.info(msg, **self.extra)
                else:
                    msg = "up_cnt higher, it's a BUG! {n}".format(n=up_cnt)
                    self.log.error(msg, priority=70, **self.extra)
                down_cnt = 0
            else:
                if down_cnt == (self.config['check_fail'] - 1):
                    self.log.info("status DOWN", priority=100, status='down',
                                  **self.extra)
                    # Service exceeded all consecutive checks.
                    # Set its state accordingly and put an item in queue.
                    # But do it only if previous state was different, to
                    # prevent unnecessary bird reloads when a service flaps
                    # between states
                    if check_state != 'DOWN':
                        check_state = 'DOWN'
                        del_operation = DeleteOperation(
                            name=self.name,
                            ip_prefix=self.config['ip_prefix'],
                            log=self.log, **self.extra)
                        msg = "{i} in queue".format(i=self.config['ip_prefix'])
                        self.log.info(msg, **self.extra)
                        self.action.put(del_operation)
                elif down_cnt < self.config['check_fail']:
                    down_cnt += 1
                    msg = "going down {n}".format(n=down_cnt)
                    self.log.info(msg, priority=40, **self.extra)
                else:
                    msg = "down_cnt higher, it's a BUG! {n}".format(n=down_cnt)
                    self.log.error(msg, priority=70, **self.extra)
                up_cnt = 0

            msg = ("wall clock time {t:.3f}ms"
                   .format(t=(time.time() - timestamp) * 1000))
            self.log.info(msg, json_blob=False)

            # calculate sleep time
            sleep = start_offset - time.time() % interval
            if sleep < 0:
                sleep += interval
            self.log.debug("sleep for {t:.3f}secs".format(t=sleep),
                           json_blob=False)
            time.sleep(sleep)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
"""
anycast healthchecker daemon
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A healthchecker for Anycasted services.

"""
__title__ = 'anycast_healthchecker'
__author__ = 'Pavlos Parissis'
__license__ = 'Apache 2.0'
__version__ = '0.5.10'
__copyright__ = 'Copyright 2015-2016 Pavlos Parissis'



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# pylint: disable=superfluous-parens
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-few-public-methods
"""
anycast_healthchecker.utils
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides utility functions and classes that are used within
anycast_healthchecker.
"""

import re
import os
import sys
import subprocess
import logging
import configparser
import shlex


OPTIONS_TYPE = {
    'check_cmd': 'get',
    'check_interval': 'getint',
    'check_timeout': 'getint',
    'check_rise': 'getint',
    'check_fail': 'getint',
    'check_disabled': 'getboolean',
    'on_disabled': 'get',
    'ip_prefix': 'get',
    'interface': 'get',
}


def valid_ip_prefix(ip_prefix):
    """Returns true if input is a valid IP Prefix otherwhise False."""

    pattern = re.compile(r'''
        ^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}  # 1-3 octets
        (?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)            # 4th octet
        (\/([0-9]|[1-2][0-9]|3[0-2]))$                      # prefixlen
        ''', re.VERBOSE)

    return pattern.match(ip_prefix)


def touch(file_path):
    """Touches a file in the same way as touch tool does"""

    try:
        with open(file_path, 'a') as file_hanle:
            os.utime(file_path, None)
    except OSError as error:
        print("Failed to touch file with error:{err}".format(err=error))
        return False
    else:
        file_hanle.close()
        return True


def get_ip_prefixes(config, services):
    """Builds a set of IP prefixes found in configuration files.

    Arguments:
        config (obg): A configparser object which holds our configuration.
        services (list): A list of section names which are the name of the
        service checks.

    Returns:
        A set of IP prefixes.
    """

    ip_prefixes = set()

    for service in services:
        ip_prefixes.add(config[service]['ip_prefix'])

    return ip_prefixes


def service_configuration_check(config, services):
    """Performs a sanity check on options for each service check

    Arguments:
        config (obj): A configparser object which holds our configuration.
        services (list): A list of section names which are the name of the
        services to run checks.

    Returns:
        None if all sanity checks on configuration are successfully passed
        otherwise exits main program
    """
    for service in services:
        for option, getter in OPTIONS_TYPE.items():
            try:
                getattr(config, getter)(service, option)
            except configparser.Error as error:
                sys.exit(str(error))
            except ValueError as error:
                sys.exit("invalid data for '{opt}' option in service check "
                         "{name}: {err}".format(opt=option, name=service,
                                                err=error))

        if (config[service]['on_disabled'] != 'withdraw' and
                config[service]['on_disabled'] != 'advertise'):
            sys.exit("'on_disable' option has invalid value ({val}) for "
                     "service check {name} should be either 'withdraw' or "
                     "'advertise'".format(name=service,
                                          val=config[service]['on_disabled']))

        # check if it is a valid IP
        if not valid_ip_prefix(config[service]['ip_prefix']):
            sys.exit("invalid value ({val}) for 'ip_prefix' option in service "
                     "check {name}. It should be an IP PREFIX in form of "
                     "ip/prefixlen.".format(name=service,
                                            val=config[service]['ip_prefix']))

        cmd = shlex.split(config[service]['check_cmd'])
        try:
            proc = subprocess.Popen(cmd, stdin=None, stdout=None, stderr=None)
            proc.kill()
        except (OSError, subprocess.SubprocessError) as error:
            sys.exit("failed to run check command '{cmd}' for service check "
                     "{name}: {err}".format(name=service,
                                            cmd=config[service]['check_cmd'],
                                            err=str(error)))


def ip_prefixes_without_config(ip_prefixes_in_bird, config, services):
    """Find IP prefixes in Bird configuration without a check.

    Arguments:
        ip_prefixes_in_bird (list): A list of IP prefixes configured in Bird.
        config (obg): A configparser object which holds our configuration.
        services (list): A list of section names which are the name of the
        service checks.

    Returns:
        A sequence (set) with IP prefixes without a check associated with them.
    """
    configured = get_ip_prefixes(config, services)
    # dummy_ip_prefix doesn't have a config by design
    configured.add(config['daemon']['dummy_ip_prefix'])

    return set(ip_prefixes_in_bird).difference(configured)


def ip_prefixes_check(config, services):
    """Reports issues with IP prefixes.

    - Report IP prefixes found in Bird configuration for which we don't have
    a service check associated with.
    - Report missing ``dummy_ip_prefix`` in Bird configuration

    Arguments:
        config (obg): A configparser object which holds our configuration.
        services (list): A list of section names which are the name of the
        service checks.
    """
    ip_prefixes_in_bird = get_ip_prefixes_from_bird(
        config['daemon']['bird_conf'])

    if not ip_prefixes_in_bird:
        print("Found zero IP prefixes in {}".format(
            config['daemon']['bird_conf']))
        return None

    notconfigured_ip_prefixes = ip_prefixes_without_config(ip_prefixes_in_bird,
                                                           config,
                                                           services)
    if notconfigured_ip_prefixes:
        print("Found IP prefixes {i} in {fh} without a check configured"
              .format(fh=config['daemon']['bird_conf'],
                      i=','.join(notconfigured_ip_prefixes)))

    if config['daemon']['dummy_ip_prefix'] not in ip_prefixes_in_bird:
        print("Dummy IP prefix ({ip}) is missing from bird configuration "
              "{fh}".format(ip=config['daemon']['dummy_ip_prefix'],
                            fh=config['daemon']['bird_conf']))


def configuration_check(config):
    """Perform a sanity check on configuration

    Arguments:
        config (obg): A configparser object which holds our configuration.

    Returns:
        None if all checks are successfully passed otherwise exits main
        program
    """
    num_level = getattr(logging, config['daemon']['loglevel'].upper(), None)
    if not isinstance(num_level, int):
        sys.exit('Invalid log level: {}'.format(config['daemon']['loglevel']))

    if not touch(config['daemon']['log_file']):
        sys.exit(1)
    if not touch(config['daemon']['stdout_file']):
        sys.exit(1)
    if not touch(config['daemon']['stderr_file']):
        sys.exit(1)
    if not valid_ip_prefix(config['daemon']['dummy_ip_prefix']):
        sys.exit("Invalid dummy IP prefix:{}".format(
            config['daemon']['dummy_ip_prefix']))

    services = config.sections()
    services.remove('daemon')
    if not services:
        sys.exit('No service checks are configured')

    service_configuration_check(config, services)
    ip_prefixes_check(config, services)


def running(processid):
    """Checks the validity of a process ID.

    Arguments:
        processid (int): Process ID number.

    Returns:
        True if process ID is found otherwise False.
    """
    try:
        # From kill(2)
        # If sig is 0 (the null signal), error checking is performed but no
        # signal is actually sent. The null signal can be used to check the
        # validity of pid
        os.kill(processid, 0)
    except OSError:
        return False

    return True


def get_ip_prefixes_from_bird(filename, die=True):
    """Builds a list of IP prefixes found in Bird configuration

    Arguments:
        filename(str): The absolute path of the Bird configuration file.

    Notes:
        It can only parse a file with the following format

            define ACAST_PS_ADVERTISE =
                [
                    10.189.200.155/32,
                    10.189.200.255/32
                ];

    Returns:
        A list of IP prefixes.
    """
    prefixes = []
    try:
        with open(filename, 'r') as bird_conf:
            lines = bird_conf.read()
    except OSError as error:
        if die:
            sys.exit(str(error))
        else:
            raise
    else:
        for line in lines.splitlines():
            line = line.strip(', ')
            if valid_ip_prefix(line):
                prefixes.append(line.rstrip(','))

    return prefixes


class BaseOperation(object):
    """Runs operation on a list


    Arguments:
        name(string): The name of the service for the given ip_prefix
        ip_prefix(string): The value to run the operation
        log(logger obj): A logger object to use for emitting messages
        extra(dictinary): A possible dictinary structure to pass further
    """
    def __init__(self, name, ip_prefix, log, **extra):
        self.name = name
        self.ip_prefix = ip_prefix
        self.log = log
        self.extra = extra


class AddOperation(BaseOperation):
    """Adds a value to a list"""
    def __str__(self):
        """Overwrite the behavior so the class

        A handy way to pass the instantiated object to a string formatter
        """
        return 'add to'

    def update(self, prefixes):
        """Add a value to the list

        Arguments:
            prefixes(list): A list to add the value
        """
        if self.ip_prefix not in prefixes:
            prefixes.append(self.ip_prefix)
            msg = ("announcing {i} for {n}".format(i=self.ip_prefix,
                                                   n=self.name))
            self.log.info(msg, **self.extra)
            return True

        return False


class DeleteOperation(BaseOperation):
    """Removes a value to a list"""
    def __str__(self):
        return 'delete from'

    def update(self, prefixes):
        """Remove a value to the list

        Arguments:
            prefixes(list): A list to remove the value
        """
        if self.ip_prefix in prefixes:
            prefixes.remove(self.ip_prefix)
            msg = "withdrawing {i} for {n}".format(i=self.ip_prefix,
                                                   n=self.name)
            self.log.info(msg, **self.extra)
            return True

        return False



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-statements
# pylint: disable=superfluous-parens
#
"""A simple healthchecker for Anycasted services.

Usage:
    anycast-healthchecker [-f <file> -d <directory> -c ] [-p | -P]

Options:
    -f, --file <file>  configuration file with settings for the daemon
                       [default: /etc/anycast-healthchecker.conf]
    -d, --dir <dir>    directory with configuration files for service checks
                       [default: /etc/anycast-healthchecker.d]
    -c, --check        perform a sanity check on configuration
    -p, --print        show default settings for daemon and service checks
    -P, --print-conf   show configuration
    -v, --version      show version
    -h, --help         show this screen
"""
import os
import sys
import signal
import logging
import configparser
import glob
import copy
from lockfile.pidlockfile import PIDLockFile
from docopt import docopt

from anycast_healthchecker import healthchecker
from anycast_healthchecker import lib
from anycast_healthchecker import __version__ as version
from anycast_healthchecker.utils import configuration_check, running

DEFAULT_OPTIONS = {
    'DEFAULT': {
        'interface': 'lo',
        'check_interval': '10',
        'check_timeout': '2',
        'check_rise': '2',
        'check_fail': '2',
        'check_disabled': 'true',
        'on_disable': 'withdraw',
    },
    'daemon': {
        'pidfile': '/var/run/anycast-healthchecker/anycast-healthchecker.pid',
        'bird_conf': '/etc/bird.d/anycast-prefixes.conf',
        'bird_variable': 'ACAST_PS_ADVERTISE',
        'loglevel': 'debug',
        'log_maxbytes': '104857600',
        'log_backups': '8',
        'log_file': '/var/log/anycast-healthchecker/anycast-healthchecker.log',
        'stderr_file': '/var/log/anycast-healthchecker/stderr.log',
        'stdout_file': '/var/log/anycast-healthchecker/stdout.log',
        'dummy_ip_prefix': '10.189.200.255/32',
        'bird_reconfigure_cmd': 'sudo /usr/sbin/birdc configure',
    }
}


def main():
    """Parse CLI and starts daemon."""

    args = docopt(__doc__, version=version)
    if args['--print']:
        for section in DEFAULT_OPTIONS:
            print("[{}]".format(section))
            for key, value in DEFAULT_OPTIONS[section].items():
                print("{k} = {v}".format(k=key, v=value))
            print()
        sys.exit(0)

    # Parse configuration.
    defaults = copy.copy(DEFAULT_OPTIONS['DEFAULT'])
    daemon_defaults = {'daemon': copy.copy(DEFAULT_OPTIONS['daemon'])}
    config = configparser.ConfigParser(defaults=defaults)
    config.read_dict(daemon_defaults)
    config_files = [args['--file']]
    config_files.extend(glob.glob(os.path.join(args['--dir'], '*.conf')))
    config.read(config_files)

    if args['--print-conf']:
        for section in config:
            print("[{}]".format(section))
            for key, value in config[section].items():
                print("{k} = {v}".format(k=key, v=value))
            print()
        sys.exit(0)

    # Perform a sanity check on the configuration
    configuration_check(config)
    if args['--check']:
        print("OK")
        sys.exit(0)

    # Catch already running process and clean up stale pid file.
    pidfile = config['daemon']['pidfile']
    if os.path.exists(pidfile):
        pid = open(pidfile).read().rstrip()
        try:
            pid = int(pid)
        except ValueError:
            print("Cleaning stale pid file with invalid data:{}".format(pid))
            os.unlink(pidfile)
        else:
            if running(pid):
                sys.exit("Process {} is already running".format(pid))
            else:
                print("Cleaning stale pid file with pid:{}".format(pid))
                os.unlink(pidfile)

    # Map log level to numeric which can be accepted by loggers.
    num_level = getattr(logging, config['daemon']['loglevel'].upper(), None)

    # Set up loggers for stdout, stderr and daemon stream
    log = lib.LoggerExt(
        'daemon',
        config['daemon']['log_file'],
        log_level=num_level,
        maxbytes=config.getint('daemon', 'log_maxbytes'),
        backupcount=config.getint('daemon', 'log_backups')
    )
    if config.getboolean('daemon', 'json_logging', fallback=False):
        log.add_central_logging(
            server=config['daemon']['http_server'],
            timeout=config.getfloat('daemon', 'http_server_timeout'),
            protocol=config['daemon']['http_server_protocol'],
            port=config['daemon']['http_server_port'])
    stdout_log = lib.LoggerExt(
        'stdout',
        config['daemon']['stdout_file'],
        log_level=num_level)

    stderrformat = ('%(asctime)s [%(process)d] line:%(lineno)d '
                    'func:%(funcName)s %(levelname)-8s %(threadName)-32s '
                    '%(message)s')
    stderr_log = lib.LoggerExt(
        'stderr',
        config['daemon']['stderr_file'],
        log_level=num_level,
        log_format=stderrformat)

    if config.getboolean('daemon', 'json_logging', fallback=False):
        stderr_log.add_central_logging(
            server=config['daemon']['http_server'],
            timeout=config.getfloat('daemon', 'http_server_timeout'),
            protocol=config['daemon']['http_server_protocol'],
            port=config['daemon']['http_server_port'])

    # Make some noise.
    log.debug('Before we are daemonized')
    stdout_log.debug('before we are daemonized')
    stderr_log.debug('before we are daemonized')

    # Get and set the DaemonContext.
    context = lib.LoggingDaemonContext()
    context.loggers_preserve = [log]
    context.stdout_logger = stdout_log
    context.stderr_logger = stderr_log

    # Set pidfile for DaemonContext
    pid_lockfile = PIDLockFile(config['daemon']['pidfile'])
    context.pidfile = pid_lockfile

    # Create our master process.
    checker = healthchecker.HealthChecker(
        log,
        config,
        config['daemon']['bird_conf'],
        config['daemon']['bird_variable'],
        config['daemon']['dummy_ip_prefix'])

    # Set signal mapping to catch singals and act accordingly.
    context.signal_map = {
        signal.SIGHUP: checker.catch_signal,
        signal.SIGTERM: checker.catch_signal,
    }

    # OK boy go and daemonize yourself.
    with context:
        log.info("starting daemon {}".format(version))
        checker.run()
# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()
